/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
This interface provides the opportunity to filter away potentially problematic input data.  It provides a test for
for both an individual sequence AND a pair of sequences (it is applied to paired-end sequences), and informs the 
user whether that sequence (or pair) passes or fails the test.  Generally, are used to eliminate reads or input contigs
with very low sequence complexity, overall low quality scores, etc.  The provision of a single test for a pair of sequences
allows for the boolean combination of the single-sequence test to be adjusted according to the test (i.e. does the pair
pass if either sequence passes, or only if both sequences pass?) and also permits the implementation of tests that apply
to the relationship of reads to one another (for example, keeping or eliminating pairs for which the two sequences extensively
overlap, indicating a short underlying amplicon).</li>
*/

#ifndef READPAIRFILTER_H
#define READPAIRFILTER_H

#include <set>
#include "ScoredSeq.h"
#include "ParameterizedInitialFile.h"
#include "ScoredSeqCollectionBwt.h"
using namespace::std;

// this is a pure virtual class that will allow the contigs generated
// at the end of each cycle to be filtered according to some arbitrary
// criteria set by the implementing classes, which are below.

class ReadPairFilter {
 public:
  virtual ~ReadPairFilter();
  virtual ReadPairFilter* copy() = 0;
  virtual bool isReadOk(ScoredSeq* read) = 0;
  virtual bool isPairOk(ScoredSeq* readA, ScoredSeq* readB) = 0;
};


// i am including the implementing classes in this file

// this one always returns true
class ReadPairFilterNull : public ReadPairFilter {
 public:
  ReadPairFilterNull();
  ~ReadPairFilterNull();
  ReadPairFilterNull* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
};

// makes sure a seq has enough score counts
class ReadPairFilterHomoPolymer : public ReadPairFilter {
 public:
  ReadPairFilterHomoPolymer();
  ReadPairFilterHomoPolymer(long maxHomoPolymer);
  ~ReadPairFilterHomoPolymer();
  ReadPairFilterHomoPolymer* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
 private:
  long _maxHomoPolymer;
};

// gets rid of sequences that match some target
class ReadPairFilterAvoidSeqs : public ReadPairFilter {
 public:
  ReadPairFilterAvoidSeqs();
  ReadPairFilterAvoidSeqs(set<ScoredSeq*>* seqsToAvoid, float minFractId);
  ~ReadPairFilterAvoidSeqs();
  ReadPairFilterAvoidSeqs* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
 private:
  ScoredSeqCollectionBwt* _seqsToAvoid;
};

// imposes multiple tests
class ReadPairFilterMulti : public ReadPairFilter {
 public:
  ReadPairFilterMulti(ReadPairFilter** testArray, int numTests);
  ~ReadPairFilterMulti();
  ReadPairFilterMulti* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
 private:
  ReadPairFilter** _testArray;
  int _numTests;
};


#endif


